/*
 * module.cpp
 *
 *  Created on: 20/02/2013
 *      Author: mmcgill
 */

#include <ros/ros.h>

#include <crosbot/config.hpp>
#include <crosbot_fastslam/module.hpp>
#include <geometry_msgs/PoseArray.h>
#include <nav_msgs/GetMap.h>

#define LOG_START                   "FastSLAMModule :: "
#define UNKNOWN_VALUE               -1

namespace crosbot {

namespace fastslam {

/**
 * Add FastSLAM to the set of map factories that can be generated by the Crosbot GUI
 */
class FastSLAMFactory {
public:
	static MapPtr factory(ConfigElementPtr config) {
		std::string type = config->getParam(PARAM_TYPE, config->getChildName());
		std::string name = config->getParam(PARAM_NAME, "");

		if (strcasecmp(type.c_str(), "fastslam") == 0) {
			ros::NodeHandle nh;
			FastSLAMModulePtr rval = new FastSLAMModule(name);
			rval->initialize(nh);
			rval->configure(config);
			return rval;
		}
		return NULL;
	}

	FastSLAMFactory() {
		MapFactory::addFactory(&factory);
	}
};
FastSLAMFactory fastSLAMFactory;
}
// namespace fastslam

} // namespace crosbot

namespace crosbot {
namespace fastslam {

class _CommandCallbackFastSLAM : public crosbot::CrosbotCommandCallback {
private:
    FastSLAMModulePtr wrapper;

public:
    _CommandCallbackFastSLAM(FastSLAMModulePtr wrapper) : wrapper(wrapper) {};
    virtual ~_CommandCallbackFastSLAM() {};
    virtual void callback_receivedCommand(const crosbot_msgs::ControlCommandPtr command) {
        if (wrapper != NULL) wrapper->callback_receivedCommand(command);
    };
};

} // namespace fastslam
} // namespace crosbot

namespace crosbot {

namespace fastslam {

FastSLAMModule::FastSLAMModule(std::string name) :
    FastSLAMMap(),
    mapFrame(DEFAULT_MAPFRAME),
    odomFrame(DEFAULT_ODOMFRAME),
    baseFrame(DEFAULT_BASEFRAME)
{
    addListener(this);
}

FastSLAMModule::~FastSLAMModule() {
    removeListener(this);
}

void FastSLAMModule::initialize(ros::NodeHandle& nh) {
    ros::NodeHandle paramNH("~");
    paramNH.param<std::string>("base_frame", baseFrame, DEFAULT_BASEFRAME);
    paramNH.param<std::string>("map_frame", mapFrame, DEFAULT_MAPFRAME);
    paramNH.param<std::string>("odom_frame", odomFrame, DEFAULT_ODOMFRAME);
    paramNH.param<std::string>("pub_history", pub_historyName, "/crosbot_fastslam/history");
    paramNH.param<std::string>("pub_map", pub_mapName, "/crosbot_fastslam/map");
    paramNH.param<std::string>("pub_particles", pub_particlesName, "/crosbot_fastslam/particles");
    paramNH.param<std::string>("srv_mapServer", srv_mapServerName, "/static_map");
    paramNH.param<std::string>("sub_poseEstimate", sub_poseEstimate, "/clicked_point");
    paramNH.param<std::string>("sub_scan", sub_scanName, "/scan");
    paramNH.param<std::string>("sub_snap", sub_snapName, "/crosbot_fastslam/snaps");
    ROS_INFO("%s map: %s, odom: %s, base: %s", LOG_START, mapFrame.c_str(), odomFrame.c_str(), baseFrame.c_str());

    // ROS subscriptions/publishers
    mapServerSrv = nh.serviceClient<nav_msgs::GetMap>(srv_mapServerName);
    scanSub = nh.subscribe(sub_scanName, 1, &FastSLAMModule::callbackScan, this);
    snapSub = nh.subscribe(sub_snapName, 1000, &FastSLAMModule::callbackSnap, this);
    gridPub = nh.advertise<nav_msgs::OccupancyGrid>(pub_mapName, 1, true);
    historyPub = nh.advertise<nav_msgs::Path>(pub_historyName, 1, true);
    particlesPub = nh.advertise<geometry_msgs::PoseArray>(pub_particlesName, 1, true);

    // Crosbot commands
    crosbotCommand = new crosbot::CrosbotCommand("", false, new _CommandCallbackFastSLAM(this));

    // read configuration/parameters
    ConfigElementPtr config = new ROSConfigElement(paramNH);
    configure(config);

    start();
}

void FastSLAMModule::shutdown() {
    stop();

    scanSub.shutdown();
    snapSub.shutdown();
    gridPub.shutdown();
    historyPub.shutdown();
}

void FastSLAMModule::callbackPoseEstimate(const geometry_msgs::PointStampedConstPtr& poseEstimate) {
    QueuedAction action;
    action.type = QueuedAction::EstimatePose;
    action.pose = Pose3D(Point3D(*poseEstimate), Quaternion());

    // Stop listening for points
    //poseEstimateSub.shutdown();

    Lock lock(incomingMutex);
    queuedActions.push(action);
    slamSemaphore.notify();

    // TODO: Add queued action for updating map again
}

void FastSLAMModule::callbackScan(const sensor_msgs::LaserScanConstPtr& latestScan) {
    Pose robotPose, sensorPose;

    tf::StampedTransform transform;
    try {
        tfListener.waitForTransform(odomFrame,
                latestScan->header.frame_id, latestScan->header.stamp,
                ros::Duration(DEFAULT_MAXWAIT4TRANSFORM));

        // Load transform laser to base
        tfListener.lookupTransform(baseFrame,
                latestScan->header.frame_id, latestScan->header.stamp, transform);
        sensorPose = transform;

        // Load transform base to odom estimate
        tfListener.lookupTransform(odomFrame, baseFrame,
                latestScan->header.stamp, transform);
        robotPose = transform;
    } catch (tf::TransformException& ex) {
        ROS_WARN("%s Problem obtaining transform (%s)\n", LOG_START, ex.what());
        return;
    }

    // Load new laser scan point cloud
    PointCloud pc(latestScan);
    newPointCloud(new PointCloud(baseFrame, pc, sensorPose.toTF()), robotPose, sensorPose);
}

void FastSLAMModule::callbackSnap(const crosbot_map::SnapMsgConstPtr& snap) {
    // TODO: convert all point clouds to base frame
    addTag(new Snap(snap));
}

void FastSLAMModule::publishTransform(const Pose& robot, const Pose& odometry, const Time& time) {
    geometry_msgs::TransformStamped ts;
    ts.header.stamp = time.toROS();
    ts.header.frame_id = mapFrame;
    ts.child_frame_id = odomFrame;

    Pose pose = robot.toTF() * odometry.toTF().inverse();

    ts.transform.translation.x = pose.position.x;
    ts.transform.translation.y = pose.position.y;
    ts.transform.translation.z = pose.position.z;
    ts.transform.rotation = pose.orientation.toROS();
    tfBroadcast.sendTransform(ts);
}

void FastSLAMModule::motionTracked(MapPtr map) {
    // TODO: Update not just pose, but everything
    ParticlePtr newMean = new Particle(mean, true);
    MapCloudPtr update = newMean->getLatestUpdate();
    if (update != NULL) {
        Pose pose = newMean->getPose();
        publishTransform(newMean->getPose(), update->odometry, update->timestamp);
    }
}

void FastSLAMModule::mapUpdated(MapPtr map) {
    // Ensure mean exists
    if (mean != NULL) {
        // TODO: Periodically call/republish map and particles - not just on new scan data/map update
        ParticlePtr newMean = new Particle(mean);

        // Use common "now" timestamp
        ros::Time tNow = ros::Time::now();

        // Publish transform
        MapCloudPtr update = newMean->getLatestUpdate();
        if (update != NULL) {
            ROS_INFO("%s map updating", LOG_START);
            Pose pose = newMean->getPose();
            publishTransform(newMean->getPose(), update->odometry, update->timestamp);
        }

        // Publish Map Occupancy Grid
        if (gridPub.getNumSubscribers() > 0) {
            nav_msgs::OccupancyGridPtr grid = asOccupancyGrid(newMean);
            if (grid != NULL) {
                grid->header.frame_id = mapFrame;
                grid->header.stamp = tNow;
                gridPub.publish(grid);
            }
        }

        // Publish history of robot path
        if (historyPub.getNumSubscribers() > 0) {
            nav_msgs::PathPtr rosPath(new nav_msgs::Path());
            rosPath->header.frame_id = mapFrame;
            rosPath->header.stamp = newMean->getLatestUpdate()->timestamp.toROS();

            PathPtr path = newMean->getPath();
            rosPath->poses.resize(path->path.size());
            for (size_t i = 0; i < path->path.size(); ++i) {
                geometry_msgs::PoseStamped& pose = rosPath->poses[i];
                pose.pose = path->path[i].toROS();
                pose.header.frame_id = mapFrame;
                pose.header.stamp = (i < path->timestamps.size())?(path->timestamps[i].toROS()):ros::Time(0);
            }

            historyPub.publish(rosPath);
        }

        // Publish particles
        if (particlesPub.getNumSubscribers() > 0) {
            geometry_msgs::PoseArray particlesMsg;
            particlesMsg.header.stamp = tNow;
            particlesMsg.header.frame_id = mapFrame;
            for (Pose& pose : particlePoses) {
                particlesMsg.poses.push_back(pose.toROS());
            }
            particlesPub.publish(particlesMsg);
        }
    }
}

void FastSLAMModule::tagAdded(MapPtr map, TagPtr tag) {
    // TODO: FastSLAMModule::tagAdded()
}

void FastSLAMModule::tagChanged(MapPtr map, TagPtr tag) {
    // TODO: FastSLAMModule::tagChanged()
}

void FastSLAMModule::callback_receivedCommand(crosbot_msgs::ControlCommandPtr command) {
    if (command->command == crosbot_msgs::ControlCommand::CMD_RESET) {
        ROS_INFO("%s Crosbot command, reset requested", LOG_START);
        resetMap();
    }
}

void FastSLAMModule::resetFastSlam() {
    // First reset the map
    FastSLAMMap::resetFastSlam();

    // Check to see if there is a map server
    if (mapServerSrv.waitForExistence(ros::Duration(1.0))) {
        ROS_INFO("%s Got map server - loading existing map", LOG_START);
        Lock lock(particlesLock, true);

        // Service Call
        nav_msgs::GetMapRequest request;
        nav_msgs::GetMapResponse response;
        mapServerSrv.call(request, response);

        nav_msgs::OccupancyGrid& grid = response.map;

        // Create new particle for the loaded map
        ParticlePtr newMap = new Particle(parameters);

        // Iterate through grid data vector - track x/y co-ordinates
        // TODO: properly set x/y from map metadata origin
        // TODO: what if map metadata pose isn't "north"
        Pose gridOrigin = Pose3D(grid.info.origin);
        float halfRes = grid.info.resolution / 2.0f;
        float x = gridOrigin.position.x + halfRes;
        float y = gridOrigin.position.y + halfRes;
        ROS_INFO("%s Map origin: %s, res: %.2lf", LOG_START, gridOrigin.toString().c_str(), grid.info.resolution);
        for (int i = 0, col = 0; i != grid.data.size(); ++i) {
            // Lookup x/y
            HeightMapPtr subMap = newMap->getMap()->getMapByXY(x, y, true, false);
            HeightVal* hVal = subMap->getByXY(x, y);
            if (hVal == NULL) {
                ROS_ERROR("%s x/y co-ordinate is null - is the map large enough?", LOG_START);
            } else {
                int value = grid.data[i];
                // Set value
                if (value == UNKNOWN_VALUE) {
                } else {
                    float prob = ProbabilityTable.lodds(value / (float) 100.0);
                    // For overlapping probs
                    if (hVal->observations > 0) {
                        hVal->prob = std::max(hVal->prob, prob);
                    } else {
                        hVal->prob = prob;
                    }

                    hVal->height = parameters.defaultHeight;
                    hVal->observations = 1;

                }

                // update x/y
                ++col;
                if (col < grid.info.width) {
                    x += grid.info.resolution;
                } else {
                    col = 0;
                    x = grid.info.origin.position.x + halfRes;
                    y += grid.info.resolution;
                }
            }

            // Set mean particle
            particles.push_back(newMap);
            particlePoses.push_back(newMap->getPose());
            mean = particles[0];
            meanPose = mean->getPose();

            // Motion is NULL
            motion = NULL;
        }

        // Temporary disable of updating the map. Map will update when pose is estimated
        setUpdateMap(false);

        // Release lock now for resample
        lock.unlock();

        // TODO: Set estimated pose? Or does the pose always wait for another call?

        // Notify map change
        mapChanged();

        // Startup sub for estimating the pose
        ros::NodeHandle nh;
        poseEstimateSub = nh.subscribe(sub_poseEstimate, 1, &FastSLAMModule::callbackPoseEstimate, this);

    } else {
        ROS_INFO("%s No map server - building new map", LOG_START);
    }

    ROS_INFO("%s Done map reset", LOG_START);
}

void FastSLAMModule::estimatePoseInMap() {
    // Generate new particles for seeding possible initial poses
    // TODO: make these configurable
    ROS_INFO("%s Generating initial pose estimates", LOG_START);
    int numAngles = 10;
    int xPointsPerDim = std::ceil(std::sqrt(parameters.numberOfParticles));
    int yPointsPerDim = xPointsPerDim;
    double widthLength = 0; //grid.info.width * grid.info.resolution;
    double heightLength = 0; //grid.info.height * grid.info.resolution;
    double xPartition = widthLength / xPointsPerDim;
    double yPartition = heightLength / yPointsPerDim;
    double xHalfPartition = xPartition / 2;
    double angleInc = 2 * M_PI / numAngles;
    double x = 0; //gridOrigin.position.x + xHalfPartition;
    double y = 0; //gridOrigin.position.y + xHalfPartition;
    for (int yPt = 0; yPt != yPointsPerDim; ++yPt) {
        x = 0;//gridOrigin.position.x + xHalfPartition;
        for (int xPt = 0; xPt != xPointsPerDim; ++xPt) {
            double angle = -M_PI;
            for (int aPt = 0; aPt != numAngles; ++aPt) {
                // Generate particle
                ParticlePtr newParticle = new Particle(mean); //new Particle(newMap);
                Pose partPose(x, y, 0, angle, 0, 0);
                newParticle->setPose(partPose);
                particles.push_back(newParticle);
                particlePoses.push_back(newParticle->getPose());

                angle += angleInc;
            }

            x += xPartition;
        }

        y += yPartition;
    }
    ROS_INFO("%s \t %d pose estimates", LOG_START, (int) particles.size());

    // Resample with higher particle count (an no motion particle)
    ROS_INFO("%s Initial resampling", LOG_START);
    int resampleCount = 3;
    int actualNumParticles = parameters.numberOfParticles;
    bool actualAddMotionCopty = parameters.addMotionCopy;
    parameters.numberOfParticles = particles.size();
    parameters.addMotionCopy = false;
    MapCloudPtr cloud = latestPointCloud;
    if (cloud == NULL) {
        ROS_INFO("%s ****** cloud is null", LOG_START);
    }
    for (int i = 0; i != resampleCount; ++i) {
        ROS_INFO("%s \t Resample %d/%d", LOG_START, i + 1, resampleCount);
        resample();

        ROS_INFO("%s \t Smearing particles", LOG_START);
        moveParticles(Pose3D(), cloud, true);

        mapChanged();
    }
    //for (int i = 0; i != particles.size(); ++i) {
    //    ROS_INFO("%s, p-%d: Pose: %s", LOG_START, i, particles[i]->getPose().toString().c_str());
    //}

    // Return resampling to normal
    ROS_INFO("%s Resample down to normal particles", LOG_START);
    parameters.numberOfParticles = actualNumParticles;
    parameters.addMotionCopy = actualAddMotionCopty;
    resample();

    // Set motion particle to the best estimate
    ParticlePtr newMotion = findMeanParticle();
    motion = new Particle(newMotion);
}

} // namespace fastslam

} // namespace crosbot
